from .estimator import EvolutionaryBaggingEstimator, EvolutionaryBaggingOldestEstimator, EvolutionaryLeveragingBaggingEstimator, PipelineHelper

__all__ = [
    "EvolutionaryBaggingEstimator",
    "EvolutionaryBaggingOldestEstimator",
    "EvolutionaryLeveragingBaggingEstimator",
    "PipelineHelper"
]



"""Utilities for unit testing and sanity checking estimators."""
import copy
import functools
import inspect
import itertools
import math
import pickle
import random

import numpy as np

__all__ = ["check_estimator"]


def yield_datasets(model):
    from sklearn import datasets as sk_datasets

    from river import base, compose, datasets, preprocessing, stream, utils

    # Multi-output regression
    if utils.inspect.ismoregressor(model):

        # 1
        yield stream.iter_sklearn_dataset(sk_datasets.load_linnerud())

        # 2
        class SolarFlare:
            """One-hot encoded version of `datasets.SolarFlare"""

            def __iter__(self):
                oh = (
                             compose.SelectType(str) | preprocessing.OneHotEncoder()
                     ) + compose.SelectType(int)
                for x, y in datasets.SolarFlare().take(200):
                    yield oh.transform_one(x), y

        yield SolarFlare()

    # Regression
    elif utils.inspect.isregressor(model):
        yield datasets.TrumpApproval().take(200)

    # Multi-output classification
    if utils.inspect.ismoclassifier(model):
        yield datasets.Music().take(200)

    # Classification
    elif utils.inspect.isclassifier(model):

        yield datasets.Phishing().take(200)
        yield ((x, np.bool_(y)) for x, y in datasets.Phishing().take(200))

        # Multi-class classification
        if model._multiclass and base.tags.POSITIVE_INPUT not in model._tags:
            yield datasets.ImageSegments().take(200)


def check_learn_one(model, dataset):
    """learn_one should return the calling model and be pure."""

    klass = model.__class__

    for x, y in dataset:
        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.learn_one(x, y)

        # Check the model returns itself
        assert isinstance(model, klass)

        # Check learn_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one(classifier, dataset):
    """predict_proba_one should return a valid probability distribution and be pure."""

    if not hasattr(classifier, "predict_proba_one"):
        return

    for x, y in dataset:

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        classifier = classifier.learn_one(x, y)
        y_pred = classifier.predict_proba_one(x)

        # Check the probabilities are coherent
        assert isinstance(y_pred, dict)
        for proba in y_pred.values():
            assert 0.0 <= proba <= 1.0
        assert math.isclose(sum(y_pred.values()), 1.0)

        # Check predict_proba_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one_binary(classifier, dataset):
    """predict_proba_one should return a dict with True and False keys."""

    for x, y in dataset:
        y_pred = classifier.predict_proba_one(x)
        classifier = classifier.learn_one(x, y)
        assert set(y_pred.keys()) == {False, True}


def assert_predictions_are_close(y1, y2):
    if isinstance(y1, dict):
        for k in y1:
            assert_predictions_are_close(y1[k], y2[k])
    elif isinstance(y1, float):
        assert math.isclose(y1, y2, rel_tol=1e-06)
    else:
        assert y1 == y2


def check_shuffle_features_no_impact(model, dataset):
    """Changing the order of the features between calls should have no effect on a model."""

    from river import utils

    params = seed_params(model._get_params(), seed=42)
    model = model._set_params(params)
    shuffled = copy.deepcopy(model)

    for x, y in dataset:

        # Shuffle the features
        features = list(x.keys())
        random.shuffle(features)
        x_shuffled = {i: x[i] for i in features}

        assert x == x_shuffled  # order doesn't matter for dicts

        if utils.inspect.isclassifier(model):
            try:
                y_pred = model.predict_proba_one(x)
                y_pred_shuffled = shuffled.predict_proba_one(x_shuffled)
            except NotImplementedError:
                y_pred = model.predict_one(x)
                y_pred_shuffled = shuffled.predict_one(x_shuffled)
        else:
            y_pred = model.predict_one(x)
            y_pred_shuffled = shuffled.predict_one(x_shuffled)

        assert_predictions_are_close(y_pred, y_pred_shuffled)

        model.learn_one(x, y)
        shuffled.learn_one(x_shuffled, y)


def check_emerging_features(model, dataset):
    """The model should work fine when new features appear."""

    for x, y in dataset:
        features = list(x.keys())
        random.shuffle(features)
        model.predict_one(x)
        model.learn_one(
            {i: x[i] for i in features[:-3]}, y
        )  # drop 3 features at random


def check_disappearing_features(model, dataset):
    """The model should work fine when features disappear."""

    for x, y in dataset:
        features = list(x.keys())
        random.shuffle(features)
        model.predict_one({i: x[i] for i in features[:-3]})  # drop 3 features at random
        model.learn_one(x, y)


def check_debug_one(model, dataset):
    for x, y in dataset:
        model.debug_one(x)
        model.learn_one(x, y)
        model.debug_one(x)
        break


def check_pickling(model, dataset):
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)
    for x, y in dataset:
        model.predict_one(x)
        model.learn_one(x, y)
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)


def check_has_tag(model, tag):
    assert tag in model._tags


def check_repr(model):
    assert isinstance(repr(model), str)


def check_str(model):
    assert isinstance(str(model), str)


def check_tags(model):
    """Checks that the `_tags` property works."""
    assert isinstance(model._tags, set)


def check_set_params_idempotent(model):
    assert len(model.__dict__) == len(model._set_params().__dict__)


def check_init_has_default_params_for_tests(model):
    for params in model._unit_test_params():
        assert isinstance(model.__class__(**params), model.__class__)


def check_init_default_params_are_not_mutable(model):
    """Mutable parameters in signatures are discouraged, as explained in
    https://docs.python-guide.org/writing/gotchas/#mutable-default-arguments
    We enforce immutable parameters by only allowing a certain list of basic types.
    """

    allowed = (type(None), float, int, tuple, str, bool, type)

    for param in inspect.signature(model.__class__).parameters.values():
        assert param.default is inspect._empty or isinstance(param.default, allowed)


def check_doc(model):
    assert model.__doc__


def check_clone(model):
    clone = model.clone()
    assert id(clone) != id(model)
    assert dir(clone) == dir(model)


def check_model_selection_order_does_not_matter(model, dataset):
    best_params = []
    permutations = list(itertools.permutations(model.models))
    datasets = itertools.tee(dataset, len(permutations))

    for permutation, dataset in zip(permutations, datasets):
        models = [model.clone() for model in permutation]
        clone = model._set_params(new_params={"models": models})
        for x, y in dataset:
            clone.predict_one(x)
            clone.learn_one(x, y)
        best_params.append(clone.best_model._get_params())

    # Check that the best params are always the same
    assert all(params == best_params[0] for params in best_params)


def seed_params(params, seed):
    """Looks for "seed" keys and sets the value."""

    def is_class_param(param):
        return (
                isinstance(param, tuple)
                and inspect.isclass(param[0])
                and isinstance(param[1], dict)
        )

    if is_class_param(params):
        return params[0], seed_params(params[1], seed)

    if not isinstance(params, dict):
        return params

    return {
        name: seed if name == "seed" else seed_params(param, seed)
        for name, param in params.items()
    }


def check_seeding_is_idempotent(model, dataset):
    params = model._get_params()
    seeded_params = seed_params(params, seed=42)

    A = model._set_params(seeded_params)
    B = model._set_params(seeded_params)

    for x, y in dataset:
        assert A.predict_one(x) == B.predict_one(x)
        A.learn_one(x, y)
        B.learn_one(x, y)


def check_multiclass_is_bool(model):
    assert isinstance(model._multiclass, bool)


def wrapped_partial(func, *args, **kwargs):
    """
    Taken from http://louistiao.me/posts/adding-__name__-and-__doc__-attributes-to-functoolspartial-objects/
    """
    partial = functools.partial(func, *args, **kwargs)
    functools.update_wrapper(partial, func)
    return partial


def allow_exception(func, exception):
    def f(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except exception:
            pass

    f.__name__ = func.__name__
    return f


def yield_checks(model):
    """Generates unit tests for a given model.
    Parameters:
        model (base.Estimator)
    """

    from river import model_selection, utils

    # General checks
    yield check_repr
    yield check_str
    yield check_tags
    yield check_set_params_idempotent
    yield check_init_has_default_params_for_tests
    yield check_init_default_params_are_not_mutable
    yield check_doc
    yield check_clone

    if utils.inspect.isclassifier(model):
        yield check_multiclass_is_bool

    # Checks that make use of datasets
    checks = [
        check_learn_one,
        check_pickling,
        check_shuffle_features_no_impact,
        check_emerging_features,
        check_disappearing_features,
    ]

    # Classifier checks
    if utils.inspect.isclassifier(model) and not utils.inspect.ismoclassifier(model):
        checks.append(allow_exception(check_predict_proba_one, NotImplementedError))
        # Specific checks for binary classifiers
        if not model._multiclass:
            checks.append(
                allow_exception(check_predict_proba_one_binary, NotImplementedError)
            )

    if isinstance(utils.inspect.extract_relevant(model), model_selection.ModelSelector):
        checks.append(check_model_selection_order_does_not_matter)

    for check in checks:
        for dataset in yield_datasets(model):
            yield wrapped_partial(check, dataset=dataset)


def check_estimator(model):
    """Check if a model adheres to `river`'s conventions.
    This will run a series of unit tests. The nature of the unit tests depends on the type of
    model.
    Parameters
    ----------
    model
    """
    for check in yield_checks(model):
        if check.__name__ in model._unit_test_skips():
            continue
        check(copy.deepcopy(model))




from river import datasets
from river import metrics
from river.evaluate import Track


def trump_mse_track(n_samples=10_000, seed=42):
    dataset = datasets.TrumpApproval().take(n_samples)
    track = Track("TRUMP Approval + R2", dataset, metrics.R2(), n_samples)
    return track

def chickweights_mse_track(n_samples=10_000, seed=42):
    dataset = datasets.ChickWeights().take(n_samples)
    track = Track("ChickWeights + R2", dataset, metrics.R2(), n_samples)
    return track



from river import compose, preprocessing, tree, linear_model, naive_bayes, neighbors, optim

from EvOAutoML.pipelinehelper import PipelineHelperRegressor
from EvOAutoML.pipelinehelper import PipelineHelperTransformer, PipelineHelperClassifier
from EvOAutoML.tracks.classification_tracks import *
from EvOAutoML.tracks.regression_tracks import *

POPULATION_SIZE = 10
N_SAMPLES = 1_000_000
N_CHECKPOINTS = 1000
SAMPLING_RATE = 1000

ENSEMBLE_CLASSIFIER = tree.HoeffdingTreeClassifier
ENSEMBLE_REGRESSOR = linear_model.LinearRegression

CLASSIFICATION_TRACKS = [
    ##('Poker Hand', pokerhand_accuracy_track),

    ('RBF(50,0.001)', rbf_accuracy_50_001_track),
    ('RBF(10,0.0001)', rbf_accuracy_10_0001_track),
    ('RBF(10,0.001)', rbf_accuracy_10_001_track),
    ('RBF(50,0.001)', rbf_accuracy_50_0001_track),

    ('SEA(50)', sea_accuracy_50_track),
    ('SEA(50000)', sea_accuracy_50000_track),

    ('Agrawal(50)', agrawal_accuracy_50_track),
    ('Agrawal(50000)', agrawal_accuracy_50000_track),

    ('LED()', led_accuracy_track),

    ('Hyperplane(50,0.001)', hyperplane_accuracy_001_track),
    ('Hyperplane(50,0.0001)', hyperplane_accuracy_0001_track),
    ('Sine', sine_accuracy_track),

    ('Covertype', covtype_accuracy_track),
    ('ELEC2', elec2_accuracy_track),
]

REGRESSION_TRACKS = [
    ('Trump Approval', trump_mse_track),
    # ('Bikes', bikes_mse_track),
    ('Chick Weights', chickweights_mse_track),
    # ('Movielens', movielens_mse_track),
    # ('Restaurants', restaurants_mse_track),
    # ('Taxi', taxis_mse_track)
]

AUTOML_CLASSIFICATION_PIPELINE = compose.Pipeline(
    ('Scaler', PipelineHelperTransformer([
        ('StandardScaler', preprocessing.StandardScaler()),
        ('MinMaxScaler', preprocessing.MinMaxScaler()),
        ('MinAbsScaler', preprocessing.MaxAbsScaler()),
        # todo create dummy
        # ('RobustScaler', preprocessing.RobustScaler()),
        # ('AdaptiveStandardScaler', preprocessing.AdaptiveStandardScaler()),
        # ('LDA', preprocessing.LDA()),

    ])),
    # ('FeatureExtractor', PipelineHelperTransformer([
    #    ('PolynomialExtender', feature_extraction.PolynomialExtender()),
    # ('RBF', feature_extraction.RBFSampler()),
    # ])),
    ('Classifier', PipelineHelperClassifier([
        ('HT', tree.HoeffdingTreeClassifier()),
        # ('FT', tree.ExtremelyFastDecisionTreeClassifier()),
        ('LR', linear_model.LogisticRegression()),
        # ('HAT', tree.HoeffdingAdaptiveTreeClassifier()),
        ('GNB', naive_bayes.GaussianNB()),
        # ('MNB', naive_bayes.MultinomialNB()),
        # ('PAC', linear_model.PAClassifier()),
        # ('ARF', ensemble.AdaptiveRandomForestClassifier()),
        ('KNN', neighbors.KNNClassifier()),
    ]))
)

CLASSIFICATION_PARAM_GRID = {
    'Scaler': AUTOML_CLASSIFICATION_PIPELINE.steps['Scaler'].generate({}),
    # 'FeatureExtractor' : AUTOML_PIPELINE.steps['FeatureExtractor'].generate({
    #    'PolynomialExtender__degree' : [1,2],
    #    'PolynomialExtender__include_bias' : [True,False],
    # 'RBF__n_components' : [2,10]
    # }),
    'Classifier': AUTOML_CLASSIFICATION_PIPELINE.steps['Classifier'].generate({
        'HT__max_depth': [10, 30, 60, 10, 30, 60],
        'HT__grace_period': [10, 100, 200, 10, 100, 200],
        'HT__max_size': [5, 10],
        #'LR__loss': [optim.losses.BinaryLoss,optim.losses.CrossEntropy],
        'LR__l2': [.0,.01,.001],
        #'LR__optimizer': [optim.SGD,optim.Adam],
        'KNN__n_neighbors': [1, 5, 20],
        'KNN__window_size': [100, 500, 1000],
        'KNN__weighted': [True, False],
        'KNN__p': [1, 2],
        # 'ARF__n_models': [5,10,5,10,5,10,5,10],
    })
}

AUTOML_REGRESSION_PIPELINE = compose.Pipeline(
    ('Scaler', PipelineHelperTransformer([
        ('StandardScaler', preprocessing.StandardScaler()),
        ('MinMaxScaler', preprocessing.MinMaxScaler()),
        ('MinAbsScaler', preprocessing.MaxAbsScaler()),
        # todo create dummy
        # ('RobustScaler', preprocessing.RobustScaler()),
        # ('AdaptiveStandardScaler', preprocessing.AdaptiveStandardScaler()),
        # ('LDA', preprocessing.LDA()),

    ])),
    ('Regressor', PipelineHelperRegressor([
        ('HT', tree.HoeffdingTreeRegressor()),
        # ('FT', tree.ExtremelyFastDecisionTreeClassifier()),
        # ('LR', linear_model.LinearRegression()),
        # ('HAT', tree.HoeffdingAdaptiveTreeClassifier()),
        # ('GNB', naive_bayes.GaussianNB()),
        # ('MNB', naive_bayes.MultinomialNB()),
        # ('PAC', linear_model.PAClassifier()),
        ('KNN', neighbors.KNNRegressor()),
    ]))
)

REGRESSION_PARAM_GRID = {
    # 'Scaler': automl_pipeline.steps['Scaler'].generate({}),
    # 'FeatureExtractor' : AUTOML_PIPELINE.steps['FeatureExtractor'].generate({
    #    'PolynomialExtender__degree' : [1,2],
    #    'PolynomialExtender__include_bias' : [True,False],
    # 'RBF__n_components' : [2,10]
    # }),
    'Regressor': AUTOML_REGRESSION_PIPELINE.steps['Regressor'].generate({
        # 'HT__tie_threshold': [.01, .05, .1],
        # 'HT__max_size' : [10,50],
        'HT__binary_split': [True, False],
        'HT__max_depth': [10, 30, 60, 10, 30, 60],
        'HT__grace_period': [10, 100, 200, 10, 100, 200],
        'HT__max_size': [5, 10],
        # 'LR__loss': [optim.losses.BinaryLoss,optim.losses.CrossEntropy],
        # 'LR__l2': [.0,.01,.001],
        # 'LR__optimizer': [optim.SGD,optim.Adam],
        'KNN__n_neighbors': [1, 5, 20],
        'KNN__window_size': [100, 500, 1000],
        # 'KNN__weighted': [True, False],
        'KNN__p': [1, 2]

        # 'HAT__tie_threshold': [.01, .05, .1],
        # 'HAT__max_size' : [10,50],

        # 'FT__max_depth': [10, 20, 50],
        # 'FT__split_confidence': [1e-7],
        # 'FT__tie_threshold': [0.05],
        # 'FT__binary_split': [False],
        # 'FT__max_size': [50, 100,200],
    })
}



import statistics

from river import base

from EvOAutoML.base.estimator import EvolutionaryBaggingEstimator


class EvolutionaryBaggingRegressor(EvolutionaryBaggingEstimator, base.Regressor):

    def predict_one(self, x: dict) -> base.typing.RegTarget:
        """Averages the predictions of each regressor."""
        arr = [regressor.predict_one(x) for regressor in self]
        return statistics.mean([.0 if v is None else v for v in arr])



import collections
import typing

from river import base

from EvOAutoML.base.estimator import EvolutionaryBaggingEstimator, EvolutionaryLeveragingBaggingEstimator, \
    EvolutionaryBaggingOldestEstimator


class EvolutionaryBaggingClassifier(EvolutionaryBaggingEstimator, base.Classifier):

    def predict_proba_one(self, x):
        """Averages the predictions of each classifier."""

        y_pred = collections.Counter()
        for classifier in self:
            y_pred.update(classifier.predict_proba_one(x))

        total = sum(y_pred.values())
        if total > 0:
            return {label: proba / total for label, proba in y_pred.items()}
        return y_pred

class EvolutionaryOldestBaggingClassifier(EvolutionaryBaggingOldestEstimator, base.Classifier):

    def predict_proba_one(self, x):
        """Averages the predictions of each classifier."""

        y_pred = collections.Counter()
        for classifier in self:
            y_pred.update(classifier.predict_proba_one(x))

        total = sum(y_pred.values())
        if total > 0:
            return {label: proba / total for label, proba in y_pred.items()}
        return y_pred

class EvolutionaryLeveragingBaggingClassifer(EvolutionaryLeveragingBaggingEstimator, base.Classifier):

    def predict_proba_one(self, x):
        """Averages the predictions of each classifier."""

        y_pred = collections.Counter()
        for classifier in self:
            y_pred.update(classifier.predict_proba_one(x))

        total = sum(y_pred.values())
        if total > 0:
            return {label: proba / total for label, proba in y_pred.items()}
        return y_pred


from . import (
    classification,
    regression,
    utils,
    base
)
from .__version__ import __version__  # noqa: F401

__all__ = [
    "classification",
    "regression",
    "utils",
    "base"
]



import random
from multiprocessing.pool import ThreadPool

import numpy as np
from river import base
from river import metrics

from EvOAutoML.classification import EvolutionaryBaggingClassifier


class ThreadEvolutionaryBestClassifier(EvolutionaryBaggingClassifier):

    def __init__(self,
                 estimator: base.Estimator,
                 param_grid,
                 population_size=10,
                 sampling_size=1,
                 metric=metrics.Accuracy,
                 sampling_rate=200,
                 seed=42
                 ):
        self.estimator = estimator
        self.param_grid = param_grid
        self.population_size = population_size
        self.sampling_size = sampling_size
        self.metric = metric
        self.sampling_rate = sampling_rate

        self.i = 0

        random.seed(seed)
        np.random.seed(seed)

        self.population = []
        self.population_metrics = []
        self._initialize_population()



    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs) -> base.Estimator:
        # Create Dataset if not initialized
        # Check if population needs to be updated
        if self.i % self.sampling_rate == 0:
            idx_best, idx_worst = self._get_best_worst_estimator_index()
            child, child_metric = self._mutate_estimator(estimator=self.population[idx_best])
            del self.population[idx_worst]
            del self.population_metrics[idx_worst]
            self.population.insert(idx_worst,child)
            self.population_metrics.insert(idx_worst,child_metric)

        # Update population
        def __update_estimator(idx: int):
            y_pred = self.population[idx].predict_one(x)
            if y_pred != {} and y_pred is not None:
                self.population_metrics[idx].update(y_true=y, y_pred=y_pred)
            self.population[idx].learn_one(x=x, y=y, **kwargs)

        self.pool = ThreadPool()
        self.pool.map(__update_estimator, range(self.population_size))
        self.pool.close()
        #for idx in range(self.population_size):
        #    __update_estimator(idx)

        self.i += 1
        return self





import typing
import pandas as pd
from river import base
import numpy as np
from river.utils import dict2numpy
from tpot import TPOTClassifier


class OnlineTpotClassifer(base.Classifier):

    def __init__(self, n_training_samples, classes: list, max_time_mins: int = 15):
        self.n_training_samples = n_training_samples
        self.max_time_mins = max_time_mins
        self.training_samples_x = []
        self.training_samples_y = []
        self.estimator = None
        self.classes = classes

    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs) -> base.Classifier:
        if self.estimator is None:
            self.training_samples_x.append(x)
            self.training_samples_y.append(y)
        if len(self.training_samples_x) >= self.n_training_samples and self.estimator is None:
            x_train = np.stack([dict2numpy(i) for i in self.training_samples_x])
            self.estimator = TPOTClassifier(max_time_mins=self.max_time_mins)
            self.estimator.fit(x_train, self.training_samples_y)
            self.training_samples_y = []
            self.training_samples_x = []
        return self

    def predict_proba_one(self, x: dict) -> typing.Dict[base.typing.ClfTarget, float]:
        if self.estimator is not None:
            y_pred = self.estimator.predict_proba([list(x.values())])[0]
            return {self.classes[i]: p for i, p in enumerate(y_pred)}
        else:
            return {c: 1 / len(self.classes) for c in self.classes}

    def predict_proba_many(self, X):
        return pd.Series(self.estimator.predict_proba(X), columns=self.classes)

    @property
    def _multiclass(self):
        return True

    def learn_many(self, X, y):
        self.estimator.partial_fit(X=X.values, y=y.values, classes=self.classes)
        return self

    def predict_one(self, x):
        if self.estimator is not None:
            y_pred = self.estimator.predict([list(x.values())])[0]
            return y_pred
        else:
            return self.classes[0]

    def predict_many(self, X):
        return pd.Series(self.estimator.predict(X))


import datetime as dt
import time
import typing
from sklearn import datasets as sk_datasets
from river import metrics
from river import utils, stream
from river.base.typing import Stream
from river.datasets import synth, Elec2, ImageSegments
from river.evaluate import Track
from river.metrics import Accuracy

from EvOAutoML.classification import EvolutionaryBaggingClassifier


class EvoTrack(Track):

    def run(self, model, n_checkpoints=10):
        # Do the checkpoint logic
        step = self.n_samples // n_checkpoints
        checkpoints = range(0, self.n_samples, step)
        checkpoints = list(checkpoints)[1:] + [self.n_samples]

        population_size = model.population_size

        # A model might be used in multiple tracks. It's a sane idea to keep things pure and clone
        # the model so that there's no side effects.
        model = model.clone()

        yield from _progressive_evo_validation(
            dataset=self.dataset,
            model=model,
            metric=self.metric,
            population_metrics=[Accuracy() for _ in range(population_size)],
            checkpoints=iter(checkpoints),
            measure_time=True,
            measure_memory=True,
        )


def _progressive_evo_validation(
    dataset: Stream,
    model: EvolutionaryBaggingClassifier,
    population_metrics,
    metric: metrics.Metric,
    checkpoints: typing.Iterator[int],
    moment: typing.Union[str, typing.Callable] = None,
    delay: typing.Union[str, int, dt.timedelta, typing.Callable] = None,
    measure_time=False,
    measure_memory=False,
):

    # Check that the model and the metric are in accordance
    if not metric.works_with(model):
        raise ValueError(
            f"{metric.__class__.__name__} metric is not compatible with {model}"
        )

    # Determine if predict_one or predict_proba_one should be used in case of a classifier
    pred_func = model.predict_one
    if utils.inspect.isclassifier(model) and not metric.requires_labels:
        pred_func = model.predict_proba_one

    preds = {}
    population_preds = {}
    for idx, metric in enumerate(population_metrics):
        population_preds[f'Individual {idx}'] = {}

    next_checkpoint = next(checkpoints, None)
    n_total_answers = 0
    if measure_time:
        start = time.perf_counter()

    for i, x, y in stream.simulate_qa(dataset, moment, delay, copy=True):

        # Question
        if y is None:
            preds[i] = pred_func(x=x)
            for idx, pred in enumerate(population_preds):
                population_preds[f'Individual {idx}'][i] = model[idx].predict_one(x)
            continue

        # Answer
        y_pred = preds.pop(i)
        y_population_pred = []
        for p in population_preds:
            t = population_preds[p].pop(i)
            y_population_pred.append(t)

        if y_pred != {} and y_pred is not None:
            metric.update(y_true=y, y_pred=y_pred)

        for idx, y_population_pred_i in enumerate(y_population_pred):
            if y_population_pred_i != {} and y_population_pred_i is not None:
                population_metrics[idx].update(y_true=y, y_pred=y_population_pred_i)

        model.learn_one(x=x, y=y)

        # Update the answer counter
        n_total_answers += 1
        if n_total_answers == next_checkpoint:
            if isinstance(metric, metrics.Metrics):
                results = {m.__class__.__name__: m.get() for m in metric}
            else:
                results = {metric.__class__.__name__: metric.get()}

            results["Accuracy"] = [results["Accuracy"]] * model.population_size
            results["Step"] = [n_total_answers] * model.population_size
            if measure_time:
                now = time.perf_counter()
                results["Time"] = [dt.timedelta(seconds=now - start)] * model.population_size
            if measure_memory:
                results["Memory"] = [model._memory_usage] * model.population_size
            results["Name"] = []
            results["Model Performance"] = []
            for idx, me in enumerate(population_metrics):
                results["Name"].append(f'Individual {idx}')
                results["Model Performance"].append(me.get())
                #results["Model Performance"].append(model.population_metrics[idx].get())

            yield results
            next_checkpoint = next(checkpoints, None)

def evo_rbf_accuracy_50_001_track(n_samples=10_000, seed=42):
    n_centroids = 50
    change_speed= .001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = EvoTrack("RBF(50,0.001)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_rbf_accuracy_10_0001_track(n_samples=10_000, seed=42):
    n_centroids = 10
    change_speed= .0001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = EvoTrack("RBF(10,0.0001)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_rbf_accuracy_10_001_track(n_samples=10_000, seed=42):
    n_centroids = 10
    change_speed= .001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = EvoTrack("RBF(10,0.001)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_rbf_accuracy_50_0001_track(n_samples=10_000, seed=42):
    n_centroids = 50
    change_speed= .0001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = EvoTrack("RBF(50,0.0001)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_sea_accuracy_50_track(n_samples=10_000, seed=42):
    width = 50
    dataset = synth.ConceptDriftStream(stream=synth.SEA(variant=1,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.SEA(variant=0,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.SEA(variant=2,seed=seed),
                                               drift_stream=synth.SEA(variant=3, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)

    track = EvoTrack("SEA(50)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_sea_accuracy_50000_track(n_samples=10_000, seed=42):
    width = 50_000
    dataset = synth.ConceptDriftStream(stream=synth.SEA(variant=1,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.SEA(variant=0,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.SEA(variant=2,seed=seed),
                                               drift_stream=synth.SEA(variant=3, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)
    track = EvoTrack("SEA(50,000)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_agrawal_accuracy_50_track(n_samples=10_000, seed=42):
    width = 50
    dataset = synth.ConceptDriftStream(stream=synth.Agrawal(classification_function=0,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.Agrawal(classification_function=2,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.Agrawal(classification_function=5,seed=seed),
                                               drift_stream=synth.Agrawal(classification_function=7, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)
    track = EvoTrack("Agrawal(50)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_agrawal_accuracy_50000_track(n_samples=10_000, seed=42):
    width = 50_000
    dataset = synth.ConceptDriftStream(stream=synth.Agrawal(classification_function=0,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.Agrawal(classification_function=2,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.Agrawal(classification_function=5,seed=seed),
                                               drift_stream=synth.Agrawal(classification_function=7, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)
    track = EvoTrack("Agrawal(50,000)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_led_accuracy_track(n_samples=10_000, seed=42):
    dataset = synth.LEDDrift(seed=seed, noise_percentage=.1, n_drift_features=4).take(n_samples)
    track = EvoTrack("LEDDrift()", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_hyperplane_accuracy_001_track(n_samples=10_000, seed=42):
    dataset = synth.Hyperplane(seed=seed,n_features=50,n_drift_features=25,mag_change=.001).take(n_samples)
    track = EvoTrack("Hyperplane(50,0.001)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_hyperplane_accuracy_0001_track(n_samples=10_000, seed=42):
    dataset = synth.Hyperplane(seed=seed,n_features=50,n_drift_features=25,mag_change=.0001).take(n_samples)
    track = EvoTrack("Hyperplane(50, 0.0001)", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_sine_accuracy_track(n_samples=10_000, seed=42):
    dataset = synth.Sine(seed=seed).take(n_samples)
    track = EvoTrack("SINE()", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_elec2_accuracy_track(n_samples=10_000, seed=42):
    dataset = Elec2().take(n_samples)
    track = EvoTrack("Elec", dataset, metrics.Accuracy(), n_samples)
    return track

def evo_covtype_accuracy_track(n_samples=10_000, seed=42):
    dataset = stream.iter_sklearn_dataset(sk_datasets.fetch_covtype())
    track = EvoTrack('Covtype', dataset, metrics.Accuracy(), n_samples)
    return track




import copy
import random
from collections import defaultdict

import numpy as np
from river import base, compose, preprocessing, tree
from river import metrics
from river.base import Estimator
from river.drift import ADWIN
from river.metrics import ClassificationMetric
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection import ParameterSampler



class EvolutionaryBaggingEstimator(base.Wrapper, base.Ensemble):

    def __init__(self, model,
                 param_grid,
                 population_size=10,
                 sampling_size=1,
                 metric=metrics.Accuracy,
                 sampling_rate=1000,
                 seed=42):

        param_iter = ParameterSampler(param_grid, population_size)
        param_list = list(param_iter)
        param_list = [dict((k, v) for (k, v) in d.items()) for d in
                      param_list]
        super().__init__([self._initialize_model(model=model,params=params) for params in param_list])
        self.param_grid = param_grid
        self.population_size = population_size
        self.sampling_size = sampling_size
        self.metric = metric
        self.sampling_rate = sampling_rate
        self.n_models = population_size
        self.model = model
        self.seed = seed
        self._rng = np.random.RandomState(seed)
        self._i = 0
        #self._drift_detectors = [copy.deepcopy(ADWIN()) for _ in range(self.n_models)]
        self._population_metrics = [copy.deepcopy(metric()) for _ in range(self.n_models)]


    @property
    def _wrapped_model(self):
        return self.model

    @classmethod
    def _unit_test_params(cls):
        model = tree.HoeffdingTreeClassifier()

        param_grid = {
            'max_depth': [10, 30, 60, 10, 30, 60],
            },

        yield {
            "model": model,
            "param_grid": param_grid,
        }

    def _initialize_model(self,model:base.Estimator,params):
        model = copy.deepcopy(model)
        model._set_params(params)
        return model

    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs):
        # Create Dataset if not initialized
        # Check if population needs to be updated
        if self._i % self.sampling_rate == 0:
            scores = [be.get() for be in self._population_metrics]
            idx_best = scores.index(max(scores))
            idx_worst = scores.index(min(scores))
            child = self._mutate_estimator(estimator=self[idx_best])
            self.models[idx_worst] = child
            self._population_metrics[idx_worst] = copy.deepcopy(self.metric())

        for idx, model in enumerate(self):
            self._population_metrics[idx].update(y_true=y, y_pred=model.predict_one(x))
            for _ in range(self._rng.poisson(6)):
                model.learn_one(x, y)
        self._i += 1
        return self


    def reset(self):
        """ Resets the estimator to its initial state.

        Returns
        -------
            self

        """
        # self.estimators = [be.reset() for be in self.estimators]
        self._i = 0
        return self

    def _mutate_estimator(self, estimator) -> (base.Classifier, ClassificationMetric):
        child_estimator = copy.deepcopy(estimator)
        key_to_change, value_to_change = random.sample(self.param_grid.items(), 1)[0]
        value_to_change = random.choice(self.param_grid[key_to_change])
        child_estimator._set_params({key_to_change: value_to_change})
        # todo refactor Mutation
        return child_estimator

    def clone(self):
        """Return a fresh estimator with the same parameters.

        The clone has the same parameters but has not been updated with any data.

        This works by looking at the parameters from the class signature. Each parameter is either

        - recursively cloned if it's a River classes.
        - deep-copied via `copy.deepcopy` if not.

        If the calling object is stochastic (i.e. it accepts a seed parameter) and has not been
        seeded, then the clone will not be idempotent. Indeed, this method's purpose if simply to
        return a new instance with the same input parameters.

        """
        return copy.deepcopy(self)


class EvolutionaryBaggingOldestEstimator(EvolutionaryBaggingEstimator):

    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs):
        # Create Dataset if not initialized
        # Check if population needs to be updated
        if self._i % self.sampling_rate == 0:
            scores = [be.get() for be in self._population_metrics]
            idx_best = scores.index(max(scores))
            idx_worst = scores.index(min(scores))
            child = self._mutate_estimator(estimator=self[idx_best])
            self.models.pop(0)
            self.models.append(child)
            self._population_metrics.pop(0)
            self._population_metrics.append(copy.deepcopy(self.metric()))

        for idx, model in enumerate(self):
            self._population_metrics[idx].update(y_true=y, y_pred=model.predict_one(x))
            for _ in range(self._rng.poisson(6)):
                model.learn_one(x, y)
        self._i += 1
        return self

class EvolutionaryLeveragingBaggingEstimator(base.Wrapper, base.Ensemble):
    """Leveraging Bagging ensemble classifier.

        Leveraging Bagging [^1] is an improvement over the Oza Bagging algorithm.
        The bagging performance is leveraged by increasing the re-sampling.
        It uses a poisson distribution to simulate the re-sampling process.
        To increase re-sampling it uses a higher `w` value of the Poisson
        distribution (agerage number of events), 6 by default, increasing the
        input space diversity, by attributing a different range of weights to the
        data samples.

        To deal with concept drift, Leveraging Bagging uses the ADWIN algorithm to
        monitor the performance of each member of the enemble If concept drift is
        detected, the worst member of the ensemble (based on the error estimation
        by ADWIN) is replaced by a new (empty) classifier.

        Parameters
        ----------
        model
            The classifier to bag.
        n_models
            The number of models in the ensemble.
        w
            Indicates the average number of events. This is the lambda parameter
            of the Poisson distribution used to compute the re-sampling weight.
        adwin_delta
            The delta parameter for the ADWIN change detector.
        bagging_method
            The bagging method to use. Can be one of the following:<br/>
            * 'bag' - Leveraging Bagging using ADWIN.<br/>
            * 'me' - Assigns $weight=1$ if sample is misclassified,
              otherwise $weight=error/(1-error)$.<br/>
            * 'half' - Use resampling without replacement for half of the instances.<br/>
            * 'wt' - Resample without taking out all instances.<br/>
            * 'subag' - Resampling without replacement.<br/>
        seed
            Random number generator seed for reproducibility.

        Examples
        --------

        >>> from river import datasets
        >>> from river import ensemble
        >>> from river import evaluate
        >>> from river import linear_model
        >>> from river import metrics
        >>> from river import optim
        >>> from river import preprocessing

        >>> dataset = datasets.Phishing()

        >>> model = ensemble.LeveragingBaggingClassifier(
        ...     model=(
        ...         preprocessing.StandardScaler() |
        ...         linear_model.LogisticRegression()
        ...     ),
        ...     n_models=3,
        ...     seed=42
        ... )

        >>> metric = metrics.F1()

        >>> evaluate.progressive_val_score(dataset, model, metric)
        F1: 0.886282

        """

    _BAGGING_METHODS = ("bag", "me", "half", "wt", "subag")

    def __init__(
            self,
            model: base.Classifier,
            param_grid,
            population_size=10,
            sampling_size=1,
            metric=metrics.Accuracy,
            sampling_rate=1000,
            w: float = 6,
            adwin_delta: float = 0.002,
            bagging_method: str = "bag",
            seed: int = None,
    ):

        param_iter = ParameterSampler(param_grid, population_size)
        param_list = list(param_iter)
        param_list = [dict((k, v) for (k, v) in d.items()) for d in
                      param_list]
        super().__init__(self._initialize_model(model=model,params=params) for params in param_list)
        self.param_grid = param_grid
        self.population_size = population_size
        self.sampling_size = sampling_size
        self.metric = metric
        self.sampling_rate = sampling_rate
        self.n_models = population_size
        self.model = model
        self.seed = seed
        self._rng = np.random.RandomState(seed)
        self._i = 0
        self._population_metrics = [copy.deepcopy(metric()) for _ in range(self.n_models)]
        self._drift_detectors = [copy.deepcopy(ADWIN(delta=adwin_delta)) for _ in range(self.n_models)]
        self.n_detected_changes = 0
        self.w = w
        self.adwin_delta = adwin_delta
        self.bagging_method = bagging_method

        # Set bagging function
        if bagging_method == "bag":
            self._bagging_fct = self._leveraging_bag
        elif bagging_method == "me":
            self._bagging_fct = self._leveraging_bag_me
        elif bagging_method == "half":
            self._bagging_fct = self._leveraging_bag_half
        elif bagging_method == "wt":
            self._bagging_fct = self._leveraging_bag_wt
        elif bagging_method == "subag":
            self._bagging_fct = self._leveraging_subag
        else:
            raise ValueError(
                f"Invalid bagging_method: {bagging_method}\n"
                f"Valid options: {self._BAGGING_METHODS}"
            )

    def _initialize_model(self,model:base.Estimator,params):
        model = copy.deepcopy(model)
        model._set_params(params)
        return model

    def _leveraging_bag(self, **kwargs):
        # Leveraging bagging
        return self._rng.poisson(self.w)

    def _leveraging_bag_me(self, **kwargs):
        # Miss-classification error using weight=1 if misclassified.
        # Otherwise using weight=error/(1-error)
        x = kwargs["x"]
        y = kwargs["y"]
        i = kwargs["model_idx"]
        error = self._drift_detectors[i].estimation
        y_pred = self.models[i].predict_one(x)
        if y_pred != y:
            k = 1
        elif error != 1.0 and self._rng.rand() < (error / (1.0 - error)):
            k = 1
        else:
            k = 0
        return k

    def _leveraging_bag_half(self, **kwargs):
        # Resampling without replacement for half of the instances
        return int(not self._rng.randint(2))

    def _leveraging_bag_wt(self, **kwargs):
        # Resampling without taking out all instances
        return 1 + self._rng.poisson(1.0)

    def _leveraging_subag(self, **kwargs):
        # Subagging using resampling without replacement
        return int(self._rng.poisson(1) > 0)

    @property
    def bagging_methods(self):
        """Valid bagging_method options."""
        return self._BAGGING_METHODS


    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs):
        # Create Dataset if not initialized
        # Check if population needs to be updated
        if self._i % self.sampling_rate == 0:
            scores = [be.get() for be in self._population_metrics]
            idx_best = scores.index(max(scores))
            idx_worst = scores.index(min(scores))
            child = self._mutate_estimator(estimator=self[idx_best])
            self.models[idx_worst] = child
            #self.population_metrics[idx_worst] = copy.deepcopy(self.metric())

        change_detected = False
        for i, model in enumerate(self):
            self._population_metrics[i].update(y_true=y, y_pred=model.predict_one(x))
            k = self._bagging_fct(x=x, y=y, model_idx=i)

            for _ in range(k):
                model.learn_one(x, y)

            y_pred = self.models[i].predict_one(x)
            if y_pred is not None:
                incorrectly_classifies = int(y_pred != y)
                error = self._drift_detectors[i].estimation
                self._drift_detectors[i].update(incorrectly_classifies)
                if self._drift_detectors[i].change_detected:
                    if self._drift_detectors[i].estimation > error:
                        change_detected = True

        if change_detected:
            self.n_detected_changes += 1
            max_error_idx = max(
                range(len(self._drift_detectors)),
                key=lambda j: self._drift_detectors[j].estimation,
            )
            self.models[max_error_idx] = copy.deepcopy(self.model)
            self._drift_detectors[max_error_idx] = ADWIN(delta=self.adwin_delta)

        return self

    def reset(self):
        """ Resets the estimator to its initial state.

        Returns
        -------
            self

        """
        # self.estimators = [be.reset() for be in self.estimators]
        self._i = 0
        return self

    def _mutate_estimator(self, estimator) -> (base.Classifier, ClassificationMetric):
        child_estimator = copy.deepcopy(estimator)
        key_to_change, value_to_change = random.sample(self.param_grid.items(), 1)[0]
        value_to_change = random.choice(self.param_grid[key_to_change])
        child_estimator._set_params({key_to_change: value_to_change})
        # todo refactor Mutation
        return child_estimator

    def clone(self):
        """Return a fresh estimator with the same parameters.

        The clone has the same parameters but has not been updated with any data.

        This works by looking at the parameters from the class signature. Each parameter is either

        - recursively cloned if it's a River classes.
        - deep-copied via `copy.deepcopy` if not.

        If the calling object is stochastic (i.e. it accepts a seed parameter) and has not been
        seeded, then the clone will not be idempotent. Indeed, this method's purpose if simply to
        return a new instance with the same input parameters.

        """
        return copy.deepcopy(self)

class PipelineHelper(Estimator):

    @classmethod
    def _unit_test_params(cls):
        models = [('HT',tree.HoeffdingTreeClassifier())]
        yield {
            "models": models,
        }

    def __init__(self, models, selected_model=None):
        self.selected_model = None
        self.models = None

        # cloned
        if type(models) == dict:
            self.models = models
        else:
            self.available_models = {}
            for (key, model) in models:
                self.available_models[key] = model

        if selected_model is None:
            self.selected_model = self.available_models[random.choice(list(self.available_models))]
        else:
            self.selected_model = selected_model



    def clone(self):
        return PipelineHelper(self.models)#, self.selected_model.clone())

    def generate(self, param_dict=None):
        if param_dict is None:
            param_dict = dict()
        per_model_parameters = defaultdict(lambda: defaultdict(list))

        # collect parameters for each specified model
        for k, values in param_dict.items():
            # example:  randomforest__n_estimators
            model_name = k.split('__')[0]
            param_name = k[len(model_name) + 2:]
            if model_name not in self.available_models:
                raise Exception('no such model: {0}'.format(model_name))
            per_model_parameters[model_name][param_name] = values

        ret = []

        # create instance for cartesion product of all available parameters
        # for each model
        for model_name, param_dict in per_model_parameters.items():
            parameter_sets = ParameterGrid(param_dict)
            for parameters in parameter_sets:
                ret.append((model_name, parameters))

        # for every model that has no specified parameters, add default value
        for model_name in self.available_models.keys():
            if model_name not in per_model_parameters:
                ret.append((model_name, dict()))

        return ret

    def _get_params(self, deep=True):
        return self.selected_model._get_params()

    def _set_params(self, new_params: dict = None):
        if len(new_params) > 0:
            self.selected_model = self.available_models[new_params[0]].__class__(**new_params[1])
        elif self.selected_model == None:
            self.selected_model = self.available_models[random.choice(list(self.available_models))]
        return self




from sklearn import datasets as sk_datasets
from river import metrics
from river.datasets import synth, Elec2, ImageSegments
from river.evaluate import Track
from river import stream


def rbf_accuracy_50_001_track(n_samples=10_000, seed=42):
    n_centroids = 50
    change_speed= .001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = Track("RBF(50,0.001)", dataset, metrics.Accuracy(), n_samples)
    return track

def rbf_accuracy_10_0001_track(n_samples=10_000, seed=42):
    n_centroids = 10
    change_speed= .0001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = Track("RBF(10,0.0001)", dataset, metrics.Accuracy(), n_samples)
    return track

def rbf_accuracy_10_001_track(n_samples=10_000, seed=42):
    n_centroids = 10
    change_speed= .001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = Track("RBF(10,0.001)", dataset, metrics.Accuracy(), n_samples)
    return track

def rbf_accuracy_50_0001_track(n_samples=10_000, seed=42):
    n_centroids = 50
    change_speed= .0001
    dataset = synth.RandomRBFDrift(seed_model=7, seed_sample=seed,n_classes=5,n_features=50, n_centroids=n_centroids, change_speed=change_speed).take(n_samples)
    track = Track("RBF(50,0.0001)", dataset, metrics.Accuracy(), n_samples)
    return track

def sea_accuracy_50_track(n_samples=10_000, seed=42):
    width = 50
    dataset = synth.ConceptDriftStream(stream=synth.SEA(variant=1,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.SEA(variant=0,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.SEA(variant=2,seed=seed),
                                               drift_stream=synth.SEA(variant=3, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)

    track = Track("SEA(50)", dataset, metrics.Accuracy(), n_samples)
    return track

def sea_accuracy_50000_track(n_samples=10_000, seed=42):
    width = 50_000
    dataset = synth.ConceptDriftStream(stream=synth.SEA(variant=1,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.SEA(variant=0,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.SEA(variant=2,seed=seed),
                                               drift_stream=synth.SEA(variant=3, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)
    track = Track("SEA(50,000)", dataset, metrics.Accuracy(), n_samples)
    return track

def agrawal_accuracy_50_track(n_samples=10_000, seed=42):
    width = 50
    dataset = synth.ConceptDriftStream(stream=synth.Agrawal(classification_function=0,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.Agrawal(classification_function=2,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.Agrawal(classification_function=5,seed=seed),
                                               drift_stream=synth.Agrawal(classification_function=7, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)
    track = Track("Agrawal(50)", dataset, metrics.Accuracy(), n_samples)
    return track

def agrawal_accuracy_50000_track(n_samples=10_000, seed=42):
    width = 50_000
    dataset = synth.ConceptDriftStream(stream=synth.Agrawal(classification_function=0,seed=seed),
                                       drift_stream=synth.ConceptDriftStream(
                                           stream=synth.Agrawal(classification_function=2,seed=seed),
                                           drift_stream=synth.ConceptDriftStream(
                                               stream=synth.Agrawal(classification_function=5,seed=seed),
                                               drift_stream=synth.Agrawal(classification_function=7, seed=seed),
                                               seed=seed,
                                               position=int(n_samples*.75),
                                               width=width
                                           ),
                                           seed=seed,
                                           position=int(n_samples*.5),
                                           width=width
                                       ),
                                       seed=seed,
                                       position=int(n_samples*.25),
                                       width=width
                                       ).take(n_samples)
    track = Track("Agrawal(50,000)", dataset, metrics.Accuracy(), n_samples)
    return track

def led_accuracy_track(n_samples=10_000, seed=42):
    dataset = synth.LEDDrift(seed=seed, noise_percentage=.1, n_drift_features=4).take(n_samples)
    track = Track("LED()", dataset, metrics.Accuracy(), n_samples)
    return track

def hyperplane_accuracy_001_track(n_samples=10_000, seed=42):
    dataset = synth.Hyperplane(seed=seed,n_features=50,n_drift_features=25,mag_change=.001).take(n_samples)
    track = Track("Hyperplane(50,0.001)", dataset, metrics.Accuracy(), n_samples)
    return track

def hyperplane_accuracy_0001_track(n_samples=10_000, seed=42):
    dataset = synth.Hyperplane(seed=seed,n_features=50,n_drift_features=25,mag_change=.0001).take(n_samples)
    track = Track("Hyperplane(50, 0.0001)", dataset, metrics.Accuracy(), n_samples)
    return track

def sine_accuracy_track(n_samples=10_000, seed=42):
    dataset = synth.Sine(seed=seed).take(n_samples)
    track = Track("SINE()", dataset, metrics.Accuracy(), n_samples)
    return track

def elec2_accuracy_track(n_samples=10_000, seed=42):
    dataset = Elec2().take(n_samples)
    track = Track("Elec", dataset, metrics.Accuracy(), n_samples)
    return track

def covtype_accuracy_track(n_samples=10_000, seed=42):
    dataset = stream.iter_sklearn_dataset(sk_datasets.fetch_covtype())
    track = Track('Covtype', dataset, metrics.Accuracy(), n_samples)
    return track

def pokerhand_accuracy_track(n_samples=10_000, seed=42):
    dataset = stream.iter_sklearn_dataset(sk_datasets.fetch_openml(data_id=354))
    track = Track('Pokerhand', dataset, metrics.Accuracy(), n_samples)
    return track




import typing

import pandas as pd
from river import base
from river.base import Estimator, Classifier, Transformer, Regressor

from EvOAutoML.base.estimator import PipelineHelper


class PipelineHelperClassifier(PipelineHelper,Classifier):

    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs) -> Estimator:
        self.selected_model = self.selected_model.learn_one(x=x, y=y, **kwargs)
        return self

    def predict_one(self, x: dict) -> base.typing.ClfTarget:
        y_pred = self.predict_proba_one(x)
        if y_pred:
            return max(y_pred, key=y_pred.get)
        return None

    def predict_proba_one(self, x: dict) -> typing.Dict[base.typing.ClfTarget, float]:
        return self.selected_model.predict_proba_one(x=x)

    def predict_proba_many(self, X: pd.DataFrame) -> pd.DataFrame:
        return self.selected_model.predict_proba_many(X=X)

    def predict_many(self, X: pd.DataFrame) -> pd.Series:
        y_pred = self.predict_proba_many(X)
        if y_pred.empty:
            return y_pred
        return y_pred.idxmax(axis="columns")

class PipelineHelperTransformer(PipelineHelper, Transformer):
    """
    Add some Text here
    """
    @property
    def _supervised(self):
        if self.selected_model._supervised:
            return True
        else:
            return False

    def transform_one(self, x: dict) -> dict:
        """

        Args:
            x:

        Returns:

        """
        return self.selected_model.transform_one(x=x)

    def learn_one(self, x: dict, y: base.typing.Target = None, **kwargs) -> "Transformer":
        """
        Add second text here
        Args:
            x:
            y:
            **kwargs:

        Returns:
            self
        """
        if self.selected_model._supervised:
            self.selected_model = self.selected_model.learn_one(x, y)
        else:
            self.selected_model = self.selected_model.learn_one(x)
        return self

class PipelineHelperRegressor(PipelineHelper, Regressor):

    def learn_one(self, x: dict, y: base.typing.ClfTarget, **kwargs) -> Estimator:
        self.selected_model = self.selected_model.learn_one(x=x, y=y, **kwargs)
        return self

    def predict_one(self, x: dict) -> base.typing.ClfTarget:
        return self.selected_model.predict_one(x=x)